-- $Id: luatexplainko-font.lua,v 1.58 2010/03/09 08:00:23 nomos Exp $
-- written by Dohyun Kim <nomos@ktug.or.kr>
-- public domain

texio.write_nl("loading luatexplainko-font.lua ...")

local initializers = fonts.initializers
local tfm	   = fonts.tfm
local define	   = fonts.define

--
-- font cache
--
---[[
local cachepath = kpse.expand_var('$TEXMFVAR') .. '/fonts/cache/luatexko'
if not lfs.isdir(cachepath) then
    local path = string.match(cachepath,"^[^/]*")
    for p in string.gmatch(cachepath,"/[^/]*") do
	path = path .. p
	if not lfs.isdir(path) then
	    lfs.mkdir(path)
	end
    end
end
luako.cachedir = file.iswritable(cachepath) and cachepath or "."

caches = caches or {}

if not md5.hex then
    function md5.hex(str)
	return string.gsub(md5.sum(str),'.',function(c) return string.format("%02x",string.byte(c)) end)
    end
end

local function cachefilename(path,name)
    return luako.cachedir .. '/' .. md5.hex(name) .. '.cache'
end

function caches.loaddata(path,name)
    local file = cachefilename(path,name)
    if lfs.isfile(file) then
	return dofile(file)
    end	
end

function caches.savedata(path,name,data)
    table.tofile(cachefilename(path,name),data,'return',false,true,false)
end

function containers.cleanname(name)
    return (string.gsub(string.lower(name),"[^%w%d\128-\255]+","-"))
end
--]]

---[[ should be deleted later
fonts.enc = fonts.enc or {} 
fonts.enc.known = fonts.enc.known or {}

local remapper = {
    otf	    = "opentype fonts",
    ttf	    = "truetype fonts",
    ttc	    = "truetype fonts",
    dfont   = "truetype fonts",		-- luatex 0.43 or up
    cid	    = "cid maps",		-- texlive2009
    fea	    = "font feature files",	-- texlive2009
}

function resolvers.find_file(name,kind)
    name = string.gsub(name,"\\","\/")
    if not kind or kind == "" then 
	kind = string.match(name,"%.(%a-)$")
    end
    if kind and kind ~= "" then
	kind = remapper[string.lower(kind)] or kind
    end
    if not kind or kind == "" then
	kind = "tex"
    end
    return kpse.find_file(name,kind)
end

function resolvers.findbinfile(name,kind)
    if not kind or kind == "" then
	kind = file.extname(name) -- string.match(name,"%.([^%.]-)$")
    end
    local filename = resolvers.find_file(name,(kind and remapper[kind]) or kind)
    if filename then return filename end
    if tex.luatexversion >= 43 then
	return resolvers.find_file(name..'.dfont','dfont')
    end
end
--]]

--
-- texlegacy, aka. tlig/trep (base only; not for node)
--
local tliglist = luako.tliglist
local treplist = luako.treplist

function initializers.common.texlegacy(tfmdata)
    -- tlig
    local chr = tfmdata.characters
    for _,v in ipairs(tliglist) do
	if chr[ v[3] ] then
	    chr[ v[1] ] = chr[ v[1] ] or {}
	    chr[ v[1] ].ligatures = chr[ v[1] ].ligatures or {}
	    chr[ v[1] ].ligatures[ v[2] ] = { char = v[3] }
	end
    end
end

table.insert(fonts.manipulators,"texlegacy")
initializers.base.otf.texlegacy = initializers.common.texlegacy
initializers.base.afm.texlegacy = initializers.common.texlegacy

local function texlegacy_trailing (fontdata)
    -- trep
    local tmps = {}
    for i,v in pairs(treplist) do
	if fontdata.characters[i] then
	    tmps[i] = fontdata.characters[v]
	    if tmps[i].ligatures and tmps[i].ligatures[v] then
		tmps[i].ligatures[i] = tmps[i].ligatures[v]
	    end
	end
    end
    for i,v in pairs(tmps) do
	fontdata.characters[i] = tmps[i]
    end
    return fontdata
end

--
-- vertical typesetting (node only; not for base)
--
function initializers.common.vertical(tfmdata)
    -- +vrt2/+vert
    local gsubs = tfmdata.luatex.features.gsub
    if gsubs then
	for k,v in pairs(gsubs) do
	    if k == "vrt2" or k == "vert" then
		tfmdata.shared.features[k] = 'yes'
		for kk,vv in pairs(v) do
		    tfmdata.script = kk
		    tfmdata.shared.features.script = kk
		    for kkk,vvv in pairs(vv) do
			tfmdata.language = kkk
			tfmdata.shared.features.language = kkk
			break
		    end
		    break
		end
		break
	    end
	end
    end
    tfmdata.direction = 8 -- TT
end

table.insert(fonts.triggers,"vertical")
initializers.node.otf.vertical = initializers.common.vertical
initializers.node.afm.vertical = initializers.common.vertical

local function vertical_trailing (fontdata)
    -- uniform height, depth
    local ht,dp 
    local factor = fontdata.vfactor or fontdata.factor
    if fontdata.ascender and fontdata.descender then
	ht = fontdata.ascender / fontdata.size
	dp = fontdata.descender / fontdata.size
    else
	ht = 0.88
	dp = 0.12
    end
    for i,v in pairs(fontdata.characters) do
	local vwidth = fontdata.descriptions and fontdata.descriptions[i] and fontdata.descriptions[i].vwidth
	if vwidth then
	    v.height = ht * vwidth * factor
	    v.depth  = dp * vwidth * factor
	else
	    v.height = ht * fontdata.size
	    v.depth  = dp * fontdata.size
	end
    end
    return fontdata
end

--
-- microtype
--
local protlist = luako.prot_list
local expnlist = luako.expand_list

function initializers.common.microtype(tfmdata)
    for i,v in pairs(tfmdata.characters) do
	v.expansion_factor = expnlist[i] and expnlist[i] or 1
    end

    for i,v in pairs(protlist) do
	local chr = tfmdata.characters[i]
	if chr then
	    chr.left_protruding  = v[1] ~= 0 and v[1]
	    chr.right_protruding = v[2] ~= 0 and v[2]
	end
    end

    tfmdata.stretch = 20
    tfmdata.shrink  = 20
    tfmdata.step    =  1
    tfmdata.auto_expand = true
end

table.insert(fonts.manipulators,"microtype")
initializers.base.otf.microtype = initializers.common.microtype
initializers.node.otf.microtype = initializers.common.microtype
initializers.base.afm.microtype = initializers.common.microtype
initializers.node.afm.microtype = initializers.common.microtype

--
-- hanging punctuations
--
local hanglist = luako.hangingpunctuations

function initializers.common.hangpunct(tfmdata)
    for i,v in pairs(hanglist) do
	local chr = tfmdata.characters[i]
	if chr then
	    chr.right_protruding = v
	end
	local dsc = tfmdata.descriptions[i]
	if dsc and dsc.lookups then
	    for _,vv in pairs(dsc.lookups) do
		for _,vvv in pairs(vv) do
		    if vvv[1] == "substitution" then
			local subst = tfmdata.luatex
			    and tfmdata.luatex.unicodes
			    and tfmdata.luatex.unicodes[ vvv[2] ]
			local substchr = tfmdata.characters[subst]
			if substchr then
			    substchr.right_protruding = chr.right_protruding
			end
		    end
		end
	    end
	end
    end

    tfmdata.auto_expand = true
end

table.insert(fonts.manipulators,"hangpunct")
initializers.base.otf.hangpunct = initializers.common.hangpunct
initializers.node.otf.hangpunct = initializers.common.hangpunct
initializers.base.afm.hangpunct = initializers.common.hangpunct
initializers.node.afm.hangpunct = initializers.common.hangpunct

--
-- non-embedding of adobe reader fonts
--

function initializers.common.noembed(tfmdata,value)
    if tex.luatexversion > 43 -- 0.44 or up
	and tfmdata.cidinfo
	and tfmdata.cidinfo.ordering
	and tfmdata.cidinfo.registry 
	and tfmdata.shared
	and tfmdata.shared.otfdata
	and tfmdata.shared.otfdata.metadata
	and tfmdata.shared.otfdata.metadata.fontname then
	    if value == "yes" then
		fonts.dontembed[tfmdata.filename] = true
	    else 
		fonts.dontembed[tfmdata.filename] = nil
	    end
    end
end

table.insert(fonts.manipulators,"noembed")
initializers.base.otf.noembed = initializers.common.noembed
initializers.node.otf.noembed = initializers.common.noembed

--
-- slant font
--
function initializers.common.fakeslant(tfmdata,value)
    if tex.luatexversion > 49 and tonumber(value) then -- 0.50 or up
	local sl = value / 1000
	tfmdata.slant_factor = sl -- luatexplain multiplies 1000
	tfmdata.has_italic = true
	local desc = tfmdata.descriptions
	for i,v in pairs(tfmdata.characters) do 
	    if desc[i] and desc[i].height and desc[i].height > 0 then
		v.italic = desc[i].height * sl
	    end
	end
    end
end

table.insert(fonts.triggers,"fakeslant")
initializers.base.otf.fakeslant = initializers.common.fakeslant
initializers.node.otf.fakeslant = initializers.common.fakeslant

--
-- extended font
--
function initializers.common.fakeextend(tfmdata,value)
    if tex.luatexversion > 49 and tonumber(value) then -- 0.50 or up
	tfmdata.extend_factor = value / 1000
    end
end

table.insert(fonts.triggers,"fakeextend")
initializers.base.otf.fakeextend = initializers.common.fakeextend
initializers.node.otf.fakeextend = initializers.common.fakeextend


--
-- virtual fonts
--

local virtualfont = {}

--
-- virtual font : underline
--
virtualfont.underline = function(f)
    local ul = f and f.shared and f.shared.features and f.shared.features.underline
    if not ul then return end
    local udown = f.descender or (0.2*f.size)
    local uwidth = 0.04*f.size
    if f.units
	and f.shared and f.shared.otfdata and f.shared.otfdata.metadata
	and f.shared.otfdata.metadata.uwidth then
	    uwidth = f.shared.otfdata.metadata.uwidth/f.units*f.size
    end
    return udown,uwidth
end

--
-- virtual font : color
--
virtualfont.colouredfont = function(f)
    local hexcolor = f and f.shared and f.shared.features and f.shared.features.color
    if not hexcolor then return end
    local red,green,blue = string.match(hexcolor,"(%x%x)(%x%x)(%x%x)")
    red   = tonumber(red,  16)/0xff
    green = tonumber(green,16)/0xff
    blue  = tonumber(blue, 16)/0xff
    return red,green,blue
end

--
-- virtualfont : fakeslant and fakeextend 
--	this should be changed to using font (or text) tags, not graphic
--
virtualfont.fakeslantextend = function(f)
    local sl = f and f.shared and f.shared.features and f.shared.features.fakeslant 
    local ex = f and f.shared and f.shared.features and f.shared.features.fakeextend
    if not sl and not ex then return f end
    sl = sl and sl/1000 or 0
    ex = ex and ex/1000 or 1
    if ex ~= 1 then
	f.parameters.quad = ex * f.parameters.quad
	f.parameters.space = ex * f.parameters.space
	f.parameters.space_stretch = ex * f.parameters.space_stretch
	f.parameters.space_shrink = ex * f.parameters.space_shrink
	f.parameters.extra_space = ex * f.parameters.extra_space
    end
    for _,v in pairs(f.characters) do
	if sl > 0 and v.height and v.height > 0 then
	    v.italic = v.height * sl
	end
	if ex ~= 1 and v.width then
	    v.width = v.width * ex
	end
    end
    return f,sl,ex
end

--
-- virtual font main func
--
virtualfont.main = function(f,id)
    -- underline
    local udown,uwidth = virtualfont.underline(f)
    -- color
    local red,green,blue = virtualfont.colouredfont(f)
    -- fakeslant/fakeextend
    local f,slant,extend = virtualfont.fakeslantextend(f)

    local vf = { 
	name = f.name,
	size = f.size,
	hash = f.hash..' @ virtual',
	parameters = f.parameters,
	type = 'virtual',
	fonts = { {id = id} },
	characters = f.characters,
	shared = f.shared,
	mode = f.mode
    }

    for i,v in pairs(vf.characters) do
	v.commands = {{'char',i}}
	if slant or udown then
	    table.insert(v.commands,1,{'push'})
	    table.insert(v.commands,{'pop'})
	end
	if slant and extend then
	    table.insert(v.commands,1,{'special',string.format("pdf:q %f 0 %f 1 0 0 cm",extend,slant)})
	    table.insert(v.commands,{'special','pdf:Q'})
	end
	if udown and uwidth then
	    table.insert(v.commands,{'down', udown})
	    table.insert(v.commands,{'rule', uwidth, v.width})
	end
	if red and green and blue then
	    table.insert(v.commands,1,{'special',string.format("pdf:%f %f %f rg %f %f %f RG",red,green,blue,red,green,blue)})
	    table.insert(v.commands,{'special','pdf:0 g 0 G'})
	end
    end

    return vf
end

--
-- main function
--
function luako.fontdefine(name,size,id)
    ---[[ xelatex's EU1 fd files
    name = string.gsub(name,"^%s*%[(.-)%]","file:%1")
    name = string.gsub(name,"%s*mapping%s*=%s*tex%-text",
	"mode=node;+kern;+liga;+dlig;featurefile=luakodefault;+tlig;+trep")
    name = string.gsub(name,"(.-):(.-):(.*)",function(a,b,c)
	if string.find(b,"italic") or string.find(b,"slant") or string.find(b,"oblique") then
	    return a..":"..b..":+itlc;"..c
	end
    end)
    --]]

    local fontdata
    if string.find(name,"[;:]%+underline")
	or string.find(name,"[;:]underline=yes")
	or string.find(name,"[;:]color=%x%x%x%x%x%x")
	or (tex.luatexversion < 50 and
		(string.find(name,"[;:]fakeslant=%d+")
		or string.find(name,"[;:]fakeextend=%d+")))
    then 
        fontdata = virtualfont.main(tfm.read_and_define(name,size))
	define.register(fontdata,font.nextid())
    else
	fontdata = define.read(name,size,id) -- by hans
    end

    if type(fontdata) == "table" then
	fontdata.userinputstring = name
	if fontdata.mode == "node" then
	    if string.find(name,"[;:]%+vertical") or string.find(name,"[;:]vertical=yes") then
		fontdata = vertical_trailing(fontdata)
	    end
	else
	    if string.find(name,"[;:]%+texlegacy") or string.find(name,"[;:]texlegacy=yes") then
		fontdata = texlegacy_trailing(fontdata)
	    end
	end
	---[[ fixed width
	if fontdata.shared
	    and fontdata.shared.otfdata
	    and fontdata.shared.otfdata.metadata
	    and fontdata.shared.otfdata.metadata.isfixedpitch then
	    fontdata.parameters.space_stretch = fontdata.parameters.space
	    if fontdata.characters[0x2423] then
		fontdata.characters[0x20] = fontdata.characters[0x2423]
	    end
	end
	--]]
    end

    return fontdata
end

callback.register('define_font', luako.fontdefine)
