-- $Id: luatexplainko-node.lua,v 1.68 2010/01/11 09:20:45 nomos Exp $
-- written by Dohyun Kim <nomos@ktug.or.kr>
-- public domain

texio.write_nl("loading luatexplainko-node.lua ...")

-- very slow when using font.fonts; instead we use fonts.ids
local fontdata		= fonts.ids or font.fonts
local get_gluenode	= nodes.glue
local get_kernnode	= nodes.kern
local get_penaltynode	= nodes.penalty
local get_rulenode	= nodes.rule

local get_hbox		= node.hpack
local remove_node	= node.remove -- nodes.remove working?
local insert_before	= node.insert_before
local insert_after	= node.insert_after
local copy_node		= node.copy
local traverse_id	= node.traverse_id
local traverse		= node.traverse
local has_attribute	= node.has_attribute
local unset_attribute	= node.unset_attribute
local set_attribute	= node.set_attribute

local cjkclass		= luako.cjkclass
local inhibitxspcode	= luako.inhibitxspcode
local postbreakpenalty	= luako.postbreakpenalty
local prebreakpenalty	= luako.prebreakpenalty
local xspcode		= luako.xspcode
local cjk_glue_spec	= luako.cjk_glue_spec
local latin_fullstop	= luako.latin_fullstop

local finemath		= luako.finemath
local cjtypeset		= luako.cjtypeset
local attrdotemph	= luako.attrdotemph
local autojosa		= luako.autojosa
local useadhocfont	= luako.useadhocfont
local attrruby		= luako.attrruby
local luakoglue		= luako.luakoglue

local kofont		= luako.kofont
local jpfont		= luako.jpfont
local csfont		= luako.csfont
local ctfont		= luako.ctfont
local setadhocfont	= luako.setadhocfont

local cjklang = { [0] = kofont, jpfont, csfont, ctfont }

local josa_list		= luako.josa_list
local josa_code		= luako.josa_code

local gluenode		= node.id("glue")
local glyphnode		= node.id("glyph")
local mathnode		= node.id("math")
local hlistnode		= node.id("hlist")
local vlistnode		= node.id("vlist")
local kernnode		= node.id("kern")
local penaltynode	= node.id("penalty")
local rulenode		= node.id("rule")
local marginkernnode	= node.id("margin_kern")
local whatsitnode	= node.id("whatsit")
local whatsitspecial	= node.subtype("special")

local emsize = 655360

local function make_luako_glue(...)
    local glue = get_gluenode(...)
    set_attribute(glue,luakoglue,1)
    return glue
end

local function is_cjk_k (c)
    return (c >= 0x2E80  and c <= 0x9FFF )
	or (c >= 0xF900  and c <= 0xFAFF )
	or (c >= 0xFE10  and c <= 0xFE1F )
	or (c >= 0xFE30  and c <= 0xFE4F )
	or (c >= 0xFF00  and c <= 0xFFEF )
	or (c >= 0x20000 and c <= 0x2A6DF)
	or (c >= 0x2F800 and c <= 0x2FA1F)
	or  c == 0x00B0  or  c == 0x2015
--	or  c == 0x2018  or  c == 0x2019
--	or  c == 0x201C  or  c == 0x201D
	or  c == 0x2026  or  c == 0x2032
	or  c == 0x2033
end

local function is_hanja (c)
    return (c >= 0x3400 and c <= 0x9FFF )
	or (c >= 0xF900 and c <= 0xFAFF )
	or (c >= 0x20000 and c <= 0x2A6DF)
	or (c >= 0x2F800 and c <= 0x2FA1F)
end

local function is_hangul (c)
    return (c >= 0xAC00 and c <= 0xD7A3)
end

local function is_chosong (c)
    return (c >= 0x1100 and c <= 0x115F)
	or (c >= 0xA960 and c <= 0xA97C)
end

local function is_jungjongsong (c)
    return (c >= 0x1160 and c <= 0x11FF)
	or (c >= 0xD7B0 and c <= 0xD7Fb)
	or c == 0x302E  or  c == 0x302F -- tone marks
end

local function get_cjk_class (ch)
    if ch then
	if is_hangul(ch) then return 7 end	  -- hangul = 7
	if is_chosong(ch) then return 8 end       -- jamo LC = 8
	if is_jungjongsong(ch) then return 9 end  -- jamo VL, TC, TM = 9
	local c = is_cjk_k(ch) and 0 or 10    -- hanja = 0; latin = 10
	if cjkclass[ch] then c = cjkclass[ch] end -- cjkclass 1 .. 6
	return c
    end
    return nil
end

local function get_font_char_data (fid,chr)
    local f = fid and (fontdata[fid] or font.fonts[fid])
    local c = chr and f and f.characters and f.characters[chr]
    return f,c
end
local function get_unicode_char(curr)
    local _,chr = get_font_char_data(curr.font,curr.char)
    if chr and chr.tounicode then
	return tonumber(chr.tounicode,16)
    else
	return curr.char
    end
end

--
-- cjk linebreaking and spacing
--

local function get_hlist_class_first (hlist)
    local curr = hlist.list
    while curr do
	if curr.id == glyphnode then
	    local c,f = get_unicode_char(curr), curr.font
	    if c then return c,f end
	elseif curr.id == hlistnode or curr.id == vlistnode then
	    local c,f = get_hlist_class_first(curr)
	    if c then return c,f end
	elseif curr.id == gluenode then
	    if curr.spec and curr.spec.width ~= 0 then return end
	end
	curr = curr.next
    end
end

local function get_hlist_class_last (hlist,prevchar,prevfont)
    local curr = node.slide(hlist.list)
    while curr do
	if curr.id == glyphnode then
	    local c,f = get_unicode_char(curr), curr.font
	    if c then return c,f end
	elseif curr.id == hlistnode or curr.id == vlistnode then
	    local c,f = get_hlist_class_last(curr)
	    if c then return c,f end
	elseif curr.id == gluenode then
	    if curr.spec and curr.spec.width ~= 0 then return end
	end
	curr = curr.prev
    end
    return prevchar, prevfont
end

local function kanjiskip (head,curr)
    insert_before(head,curr,make_luako_glue(0, emsize*0.1, emsize*0.01))
end

local function xkanjiskip (head,curr)
    if has_attribute(curr,finemath) == 0 then -- ttfamily
	kanjiskip(head,curr)
    else
	insert_before(head,curr,make_luako_glue(0.25*emsize, emsize*0.15, emsize*0.06))
    end
end

local function interhangulskip (head,curr,currfont,prevfont,was_penalty)
    local width = 0
    local interhangul = currfont and fontdata[currfont]
	and fontdata[currfont].shared
        and fontdata[currfont].shared.features
        and fontdata[currfont].shared.features.interhangul
    if tonumber(interhangul) and currfont == prevfont then
        width = (interhangul - 1000)/1000 * emsize
    end
    if not was_penalty then
	insert_before(head,curr,get_penaltynode(50))
    end
    insert_before(head,curr,make_luako_glue(width, emsize*0.01, emsize*0.01))
end

local function interhanjaskip (head,curr,was_penalty)
    if not was_penalty then
	insert_before(head,curr,get_penaltynode(50))
    end
    insert_before(head,curr,make_luako_glue(0, emsize*0.01, emsize*0.01))
end

local function koreanlatinskip (head,curr,currfont,prevfont,was_penalty)
    local width = 0
    if has_attribute(curr,finemath) == 1 then -- not ttfamily
        local latincjk = currfont
            and fontdata[currfont]
            and fontdata[currfont].shared
            and fontdata[currfont].shared.features
            and fontdata[currfont].shared.features.interlatincjk
        if not tonumber(latincjk) then
            latincjk = prevfont
                and fontdata[prevfont]
                and fontdata[prevfont].shared
                and fontdata[prevfont].shared.features
                and fontdata[prevfont].shared.features.interlatincjk
        end
        if tonumber(latincjk) then
            width = (latincjk - 1000)/1000 * emsize
        end
    end
    if not was_penalty then
	insert_before(head,curr,get_penaltynode(50))
    end
    insert_before(head,curr,make_luako_glue(width, emsize*0.01, emsize*0.01))
end

local function cjk_insert_nodes(head,curr,currchar,currfont,prevchar,prevfont)
    local was_penalty = false
    local p = get_cjk_class(prevchar)
    local c = get_cjk_class(currchar)
    if prebreakpenalty[currchar] then
	was_penalty = true
	insert_before(head,curr,get_penaltynode(prebreakpenalty[currchar]))
    elseif postbreakpenalty[prevchar] then
	was_penalty = true
	insert_before(head,curr,get_penaltynode(postbreakpenalty[prevchar]))
    end
    if p and c then
	if has_attribute(curr,cjtypeset) then
	    if cjk_glue_spec[p] and cjk_glue_spec[p][c] then
		local width   = emsize * cjk_glue_spec[p][c][1]
		local stretch = 0
		local shrink  = emsize * cjk_glue_spec[p][c][2] 
		insert_before(head,curr,make_luako_glue(width, stretch, shrink))
	    elseif p < 10 and c < 10 and p ~= 8 and c ~= 9 then
		kanjiskip(head,curr)
	    elseif (p < 10 and c == 10) or (p == 10 and c < 10) then
		if xspcode[currchar] then
		    if xspcode[currchar] % 2 == 1 then
			xkanjiskip(head,curr)
		    end
		elseif xspcode[prevchar] then
		    if xspcode[prevchar] > 1 then
			xkanjiskip(head,curr)
		    end
		elseif inhibitxspcode[currchar] then -- 3, 2
		    if inhibitxspcode[currchar] > 1 then
			xkanjiskip(head,curr)
		    end
		elseif inhibitxspcode[prevchar] then -- 3, 1
		    if inhibitxspcode[prevchar] % 2 == 1 then
			xkanjiskip(head,curr)
		    end
		else
		    xkanjiskip(head,curr)
		end
	    end
	else
	    if (p < 10 and c == 10) or (p == 10 and c < 10) then
		if xspcode[currchar] then
		    if xspcode[currchar] % 2 == 1 then
			koreanlatinskip(head,curr,currfont,prevfont,was_penalty)
		    end
		elseif xspcode[prevchar] then
		    if xspcode[prevchar] > 1 then
			koreanlatinskip(head,curr,currfont,prevfont,was_penalty)
		    end
		elseif inhibitxspcode[currchar] then -- 3, 2
		    if inhibitxspcode[currchar] > 1 then
			koreanlatinskip(head,curr,currfont,prevfont,was_penalty)
		    end
		elseif inhibitxspcode[prevchar] then -- 3, 1
		    if inhibitxspcode[prevchar] % 2 == 1 then
			koreanlatinskip(head,curr,currfont,prevfont,was_penalty)
		    end
		else
		    koreanlatinskip(head,curr,currfont,prevfont,was_penalty)
		end
	    elseif cjk_glue_spec[p] and cjk_glue_spec[p][c] then
		koreanlatinskip(head,curr,currfont,prevfont,was_penalty)
	    elseif p == 7 and c == 7 then
		interhangulskip(head,curr,currfont,prevfont,was_penalty)
	    elseif p < 10 and c < 10 and p ~= 8 and c ~= 9 then
		interhanjaskip(head,curr,was_penalty)
	    end
	end
	if curr.id == glyphnode
	    and has_attribute(curr,finemath) == 1 -- not ttfamily
	    and latin_fullstop[currchar] then 
	    if p < 10 or latin_fullstop[prevchar] then
		local punctraise = prevfont
		    and fontdata[prevfont]
		    and fontdata[prevfont].shared
		    and fontdata[prevfont].shared.features
		    and fontdata[prevfont].shared.features.punctraise
		punctraise = tonumber(punctraise) and ((punctraise-1000)/1000*emsize)
		if punctraise then
		    curr.yoffset = curr.yoffset and (curr.yoffset+punctraise) or punctraise
		end
		local latincjk = prevfont
		    and fontdata[prevfont]
		    and fontdata[prevfont].shared
		    and fontdata[prevfont].shared.features
		    and fontdata[prevfont].shared.features.interlatincjk
		latincjk = tonumber(latincjk) and ((latincjk-1000)/1000*emsize)
		if latincjk then
		    if latin_fullstop[prevchar] then 
			if curr.prev and curr.prev.prev and curr.prev.prev.id == kernnode then
			else
			    insert_before(head,curr,get_kernnode(latincjk))
			end
		    end
		    if curr.next and curr.next.id == glyphnode then
			local nextchar = get_unicode_char(curr.next)
			if nextchar then
			    if latin_fullstop[currchar] ~= 1 and (nextchar == 0x201D or nextchar == 0x2019) then
				insert_after(head,curr,get_kernnode(latincjk))
			    end
			end
		    end
		    currfont = prevfont
		end
	    end
	end
    end
    -- for dot emphasis
    if has_attribute(curr,attrdotemph)
	and c ~= 0
	and c ~= 7
	and p ~= 8
	and c ~= 8 then
	unset_attribute(curr,attrdotemph)
    end

    return currchar,currfont
end

local function cjk_spacing_linebreak (head)
    local prevchar,prevfont = nil,nil
    for curr in traverse(head) do
	if has_attribute(curr,finemath) then
	    if curr.id == gluenode then 
		prevchar,prevfont = nil,nil
		unset_attribute(curr,finemath)
	    elseif curr.id == glyphnode then
		local f = get_font_char_data(curr.font)
		emsize = (f and f.parameters and f.parameters.quad) or emsize
		local uni = get_unicode_char(curr)
		if uni then
		    prevchar,prevfont = cjk_insert_nodes(head,curr,uni,curr.font,prevchar,prevfont)
		end
		unset_attribute(curr,finemath)
	    elseif curr.id == mathnode then
		local currchar = 0
		if curr.surround and curr.surround > 0 then
		    currchar = 0x4E00
		end
		if curr.subtype == 0 then
		    cjk_insert_nodes(head,curr,currchar,nil,prevchar,prevfont)
		    prevchar,prevfont = nil,nil
		else
		    prevchar,prevfont = currchar,nil
		end
		unset_attribute(curr,finemath)
	    elseif curr.id == hlistnode or curr.id == vlistnode then
		local firstchr, firstfid = get_hlist_class_first(curr) 
		if firstchr then
		    cjk_insert_nodes(head,curr,firstchr,firstfid,prevchar,prevfont)
		end
		prevchar,prevfont = get_hlist_class_last(curr,prevchar,prevfont)
		unset_attribute(curr,finemath)
	    end
	else
	    prevchar,prevfont = 0,nil -- treat \verb as latin character.
	end
    end
end

--
-- remove japanese/chinese spaceskip
--

local function remove_cj_spaceskip (head)
    for curr in traverse_id(gluenode,head) do
	local cjattr = has_attribute(curr,cjtypeset)
	if cjattr and cjattr > 0 and curr.spec and curr.prev and curr.next then
	    local space,stretch,shrink,prevfont,prevchar,prevclass,nextclass
	    if curr.prev.id == glyphnode then
		prevclass = get_cjk_class(get_unicode_char(curr.prev))
		prevfont = curr.prev.font
		prevchar = curr.prev.char
	    elseif curr.prev.id == hlistnode or curr.prev.id == vlistnode then
		prevclass,prevfont = get_hlist_class_last(curr.prev)
	    end
	    if curr.next.id == glyphnode then
		nextclass = get_cjk_class(get_unicode_char(curr.next))
	    elseif curr.next.id == hlistnode or curr.next.id == vlistnode then
		nextclass = get_hlist_class_first(curr.next)
	    end
	    if tex.spaceskip.width == 0 then
		prevfont = get_font_char_data(prevfont)
		if prevfont and prevfont.parameters then
		    space = prevfont.parameters.space and tex.round(prevfont.parameters.space)
		    stretch = prevfont.parameters.space_stretch and tex.round(prevfont.parameters.space_stretch)
		    shrink = prevfont.parameters.space_shrink and tex.round(prevfont.parameters.space_shrink)
		end
	    else
		space   = tex.spaceskip.width
		stretch = tex.spaceskip.stretch
		shrink  = tex.spaceskip.shrink
	    end
	    if prevchar and prevchar >= 65 and prevchar <= 90 then
		-- short of satisfactory solution: ) etc
		stretch = stretch and math.floor(stretch*(999/1000))
		shrink  = shrink  and math.floor(shrink*(1001/1000))
	    end
	    if curr.spec.width == space and curr.spec.stretch == stretch and curr.spec.shrink == shrink then 
		if prevclass and prevclass < 10 then 
		    remove_node(head,curr) 
		elseif nextclass and nextclass < 10 then
		    remove_node(head,curr) 
		end
	    end
	end
    end
end

--
-- compress fullwidth punctuations
--

local function get_hss_glue ()
    local hss = get_gluenode(0,65536,65536)
    hss.spec.stretch_order = 1
    hss.spec.shrink_order  = 1
    return hss
end

local function compress_fullwidth_punctuations (head)
    for curr in traverse_id(glyphnode,head) do
	local uni = get_unicode_char(curr)
	local f,chr = get_font_char_data(curr.font,curr.char)
	if uni and chr then
	    emsize = (f and f.parameters and f.parameters.quad) or emsize
	    if cjkclass[uni] == 2 or cjkclass[uni] == 4 then
		if chr.right_protruding and chr.right_protruding >= 500 then
		    insert_after (head,curr,get_kernnode(-0.5*emsize))
		else
		    local hss = get_hss_glue()
		    local nch = copy_node(curr)
		    nch.next = hss
		    insert_before(head,curr,get_hbox(nch,emsize*0.5,"exactly"))
		    remove_node(head,curr)
		end
	    elseif cjkclass[uni] == 1 then
		if chr.left_protruding and chr.left_protruding >= 500 then
		    insert_before(head,curr,get_kernnode(-0.5*emsize))
		else
		    local hss = get_hss_glue()
		    local nch = copy_node(curr)
		    hss.next = nch
		    insert_before(head,curr,get_hbox(hss,emsize*0.5,"exactly"))
		    remove_node(head,curr)
		end
	    elseif cjkclass[uni] == 3 then 
		if (chr.left_protruding and chr.left_protruding >= 250)
		    or (chr.right_protruding and chr.right_protruding >= 250) then
		    if chr.left_protruding then 
			insert_before(head,curr,get_kernnode(-0.25*emsize))
		    end
		    if chr.right_protruding then
			insert_after (head,curr,get_kernnode(-0.25*emsize))
		    end
		else
		    local hssa = get_hss_glue()
		    local hssb = get_hss_glue()
		    local nch = copy_node(curr)
		    hssa.next = nch
		    nch.next = hssb
		    insert_before(head,curr,get_hbox(hssa,emsize*0.5,"exactly"))
		    remove_node(head,curr)
		end
	    end
	end
    end
end

--
-- ruby: pre-linebreak routine
--

local function spread_ruby_box(head,extrawidth)
    for curr in traverse_id(gluenode,head) do
	if curr.spec then 
	    if has_attribute(curr,luakoglue) then
		curr.spec.width = curr.spec.width + extrawidth
	    else
		insert_before(head,curr,get_kernnode(curr.spec.width+extrawidth))
		remove_node(head,curr)
	    end
	end
    end
end

local function spread_ruby_base_box (head)
    for curr in traverse_id(hlistnode,head) do
	local attr = has_attribute(curr,attrruby)
	local rubyoverlap = attr and luako.rubynode[attr][3]
	if attr and not rubyoverlap then
	    local ruby = luako.rubynode[attr][1]
	    if ruby and ruby.width > curr.width then
		local basehead = curr.list
		local numofglues = node.count(gluenode,basehead)
		local extrawidth = (ruby.width - curr.width)/(numofglues + 1)
		if numofglues > 0 then
		    spread_ruby_box(basehead,extrawidth)
		end
		local leading  = get_kernnode(extrawidth/2)
		curr.width = ruby.width
		leading.next = basehead
		curr.list = leading
	    end
	end
    end
end

local function get_ruby_side_width (basewidth,rubywidth,adjacent)
    local width,margin = (rubywidth-basewidth)/2, 0
    if adjacent then
	if adjacent.id == glyphnode then
	    if not is_hanja(get_unicode_char(adjacent)) then
		width = (rubywidth-basewidth-emsize)/2
		if width > 0 then
		    margin = emsize/2
		else
		    width,margin = 0,(rubywidth-basewidth)/2
		end
	    end
	end
    end
    return width,margin
end

local function get_ruby_side_kern (head)
    for curr in traverse_id(hlistnode,head) do
	local attr = has_attribute(curr,attrruby)
	local rubyoverlap = attr and luako.rubynode[attr][3]
	if rubyoverlap then
	    local basewidth = curr.width
	    local rubywidth = luako.rubynode[attr][1].width
	    if curr.width < rubywidth then
		local _,f = get_hlist_class_first(curr)
		f = get_font_char_data(f)
		emsize = (f and f.parameters and f.parameters.quad) or emsize
		local leftwidth,leftmargin = get_ruby_side_width(basewidth,rubywidth,curr.prev)
		if leftwidth > 0 then
		    curr.width = curr.width + leftwidth
		end
		if leftmargin > 0 then
		    luako.rubynode[attr].leftmargin = leftmargin
		end
		local rightwidth,rightmargin = get_ruby_side_width(basewidth,rubywidth,curr.next)
		if rightwidth > 0 then
		    curr.width = curr.width + rightwidth
		end
		if rightmargin > 0 then
		    luako.rubynode[attr].rightmargin = rightmargin
		end
		luako.rubynode[attr].rightshift = rightwidth - leftwidth
		local totalspace = leftwidth+rightwidth
		if totalspace > 0 then
		    local numofglues = node.count(gluenode,curr.list)
		    local extrawidth = totalspace/(numofglues + 1)
		    if numofglues > 0 then
			spread_ruby_box(curr.list,extrawidth)
		    end
		    local leading = get_kernnode(extrawidth*(leftwidth/totalspace))
		    leading.next = curr.list
		    curr.list = leading
		end
	    end
	end
    end
end

local function zero_width_rule_with_dir (head,curr,before)
    local rule = get_rulenode(0)
    rule.dir = curr.dir
    if before then
	insert_before(head,curr,rule)
    else
	insert_after(head,curr,rule)
    end
end

local function no_ruby_at_margin(head)
    for curr in traverse_id(hlistnode,head) do
	local attr = has_attribute(curr,attrruby)
	if attr then
	    local margin = luako.rubynode[attr].leftmargin
	    if margin then
		insert_before(head,curr,get_kernnode(-margin))
		zero_width_rule_with_dir(head,curr,true) -- before
		insert_before(head,curr,get_kernnode(margin))
	    end
	    margin = luako.rubynode[attr].rightmargin
	    if margin then
		if curr.next then
		    if curr.next.id == gluenode then
			insert_after(head,curr,get_kernnode(-margin))
			zero_width_rule_with_dir(head,curr)
			insert_after(head,curr,get_kernnode(margin))
		    elseif curr.next.id == penaltynode and curr.next.penalty < 10000 then
			insert_after(head,curr.next,get_kernnode(-margin))
			zero_width_rule_with_dir(head,curr)
			insert_after(head,curr,get_kernnode(margin))
		    end
		end
	    end
	end
    end
end

--
-- discourage character widow (고아글자 회피)
--

local function inject_char_widow_penalty (head,curr,uni,pv)
    if uni and prebreakpenalty[uni] ~= 10000 then 
	local class = get_cjk_class(uni)
	if class and class < 9 then
	    if curr.prev and curr.prev.id == gluenode then
		curr = curr.prev
	    end
	    if curr.prev and curr.prev.id == penaltynode then
		if curr.prev.penalty < pv then
		    curr.prev.penalty = pv
		end
	    else 
		insert_before(head,curr,get_penaltynode(pv))
	    end
	    return true
	end
    end
end

local function discourage_char_widow (head,curr)
    local width = 0
    while curr do
	local pv = has_attribute(curr,cjtypeset) and 500 or 5000
	if curr.id == glyphnode then
	    local f,c = get_font_char_data(curr.font,curr.char)
	    emsize = (f and f.parameters and f.parameters.quad) or emsize
	    width = width + ((c and c.width) or 0)
	    if width >= 2*emsize then return end
	    local uni = get_unicode_char(curr)
	    if inject_char_widow_penalty(head,curr,uni,pv) then return true end
	elseif curr.id == hlistnode and curr.id == vlistnode then
	    width = width + (curr.width or 0)
	    local uni,f = get_hlist_class_first(curr)
	    f = get_font_char_data(f)
	    emsize = (f and f.parameters and f.parameters.quad) or emsize
	    if width >= 2*emsize then return end
	    if inject_char_widow_penalty(head,curr,uni,pv) then return true end
	end
	curr = curr.prev
    end
end

--
-- automatic josa selection (자동조사)
--

local function syllable2jamo (code)
    local code = code - 0xAC00
    local L = 0x1100 + math.floor(code / 588)
    local V = 0x1161 + math.floor((code % 588) / 28)
    local T = 0x11A7 + code % 28
    if T == 0x11A7 then T = nil end
    return L, V, T
end

local function number2josacode (n)
    n = n % 10 + string.byte("0")
    if josa_code[n] then return josa_code[n] end
    return nil -- 2
end

local function latin2josacode (n)
    n = n + string.byte("a")
    if josa_code[n] then return josa_code[n] end
    return nil -- 2
end

for c = 0x2160, 0x216B do -- Ⅰ
    josa_code[c] = number2josacode(c - 0x215F)
end
for c = 0x2170, 0x217B do -- ⅰ
    josa_code[c] = number2josacode(c - 0x216F)
end
for c = 0x2460, 0x2473 do -- ①
    josa_code[c] = number2josacode(c - 0x245F)
end
for c = 0x2474, 0x2487 do -- ⑴
    josa_code[c] = number2josacode(c - 0x2473)
end
for c = 0x2488, 0x249B do -- ⒈
    josa_code[c] = number2josacode(c - 0x2487)
end
for c = 0x249C, 0x24B5 do -- ⒜
    josa_code[c] = latin2josacode(c - 0x249C)
end
for c = 0x24B6, 0x24CF do --  Ⓐ
    josa_code[c] = latin2josacode(c - 0x24B6)
end
for c = 0x24D0, 0x24E9 do --  ⓐ
    josa_code[c] = latin2josacode(c - 0x24D0)
end
for c = 0x3131, 0x314E do -- ㄱ
    josa_code[c] = 3 
end
josa_code[0x3139] = 1 -- ㄹ
for c = 0x3165, 0x3186 do
    josa_code[c] = 3 
end
for c = 0x3200, 0x320D do -- ㈀
    josa_code[c] = 3
end
josa_code[0x3203] = 1 -- ㈃
for c = 0x3260, 0x327F do -- ㉠
    josa_code[c] = 3
end
josa_code[0x3263] = 1 -- ㉣
for c = 0xFF10, 0xFF19 do -- ０
    josa_code[c] = number2josacode(c - 0xFF10)
end
for c = 0xFF21, 0xFF3A do -- Ａ
    josa_code[c] = latin2josacode(c - 0xFF21)
end
for c = 0xFF41, 0xFF5A do -- ａ
    josa_code[c] = latin2josacode(c - 0xFF41)
end

local function jamo2josacode(code)
    if code and code > 0x11A7 then
	if code == 0x11AF then return 1 end
	return 3
    end
    return 2
end

local function get_hanja_hangul_table (table,file,init)
    local i = 0
    local file = kpse.find_file(file)
    if not file then return table end
    file = io.open(file, "r")
    if not file then return table end
    while true do
	local d = file:read("*number")
	if not d then break end
	table[init + i] = d
	i = i + 1
    end
    file:close()
    return table
end

local hanja2hangul = { }
hanja2hangul = get_hanja_hangul_table(hanja2hangul,"hanja_hg.tab",0x4E00)
hanja2hangul = get_hanja_hangul_table(hanja2hangul,"hjexa_hg.tab",0x3400)
hanja2hangul = get_hanja_hangul_table(hanja2hangul,"hjcom_hg.tab",0xF900)

-- 1 : 리을,    2 : 중성,    3 : 종성 
local function get_josacode (prevs)
    local code = prevs[1] -- last char
    if not code then return 2 end
    if is_hangul(code) then -- hangul syllable
	local _, _, T = syllable2jamo(code)
	return jamo2josacode(T)
    end
    if is_jungjongsong(code) then return jamo2josacode(code) end
    if (code >= 0x3400 and code <= 0x9FA5)
	or (code >= 0xF900 and code <= 0xFA2D) then
	local _, _, T = syllable2jamo(hanja2hangul[code])
	return jamo2josacode(T)
    end
    -- latin
    if prevs[1] < 0x80
	and prevs[2] and prevs[2] < 0x80
	and prevs[3] and prevs[3] < 0x80 then
	local liii = string.char(prevs[3], prevs[2], prevs[1]) 
	if josa_code[liii] then return josa_code[liii] end
    end
    if prevs[1] < 0x80
	and prevs[2] and prevs[2] < 0x80 then
	local lii = string.char(prevs[2], prevs[1])   
	if josa_code[lii] then return josa_code[lii] end
    end
    if josa_code[code] then return josa_code[code] end
    return 2
end

local function get_josaprevs(curr,josaprev)
    while curr do
	if curr.id == glyphnode then
	    local chr = get_unicode_char(curr)
	    if xspcode[chr] 
		or inhibitxspcode[chr]
		or prebreakpenalty[chr]
		or postbreakpenalty[chr]
		or chr == 0x302E -- tone mark
		or chr == 0x302F -- tone mark
		then
		    --skip
	    else
		josaprev[#josaprev + 1] = chr
	    end
	elseif curr.id == hlistnode or curr.id == vlistnode then
	    josaprev = get_josaprevs(node.slide(curr.list),josaprev)
	end
	if #josaprev == 3 then break end
	curr = curr.prev
    end
    return josaprev
end

local function korean_autojosa (head)
    for curr in traverse_id(glyphnode,head) do
	if has_attribute(curr,autojosa) and has_attribute(curr,finemath) then
	    local josaprev = {}
	    josaprev = get_josaprevs(curr.prev,josaprev)
	    local josacode = get_josacode(josaprev)
	    local thischar = get_unicode_char(curr)
	    if thischar == 0xC774 then 
		if curr.next
		    and curr.next.id == glyphnode
		    and get_unicode_char(curr.next) == 0xB77C then
		    curr.char = josa_list[0xC774][josacode]
		else
		    curr.char = josa_list[0xAC00][josacode]
		end
	    elseif thischar and josa_list[thischar] then
		curr.char = josa_list[thischar][josacode]
	    end
	end
	if curr.char < 0 then remove_node(head,curr) end
    end
end

--
-- font substitution
--

local cjk_font_list = {}

local function font_substitute (head) 
    for curr in traverse_id(glyphnode,head) do
	local eng,engchr = get_font_char_data(curr.font,curr.char)
	local korid, fallback
	if eng and eng.size and curr.char and not engchr then
	    local myadhoc = has_attribute(curr,useadhocfont)
	    if myadhoc and setadhocfont[myadhoc] then
		fallback = setadhocfont[myadhoc]
	    else
		local lang = has_attribute(curr,cjtypeset) or 0
		lang = cjklang[lang] or kofont
		local family,series,shape
		local texfnt = tex.fontidentifier(curr.font)
		if texfnt then
		    family,series,shape = string.match(texfnt,"/(%a+)/(%a+)/(%a+)/")
		end
		family = (family and luako.ltfont[family]) or "rm"
		fallback = lang[family] or lang.rm
		fallback = (series and fallback[series]) or fallback.m
		fallback = (shape and fallback[shape]) or fallback.n
	    end
	    if fallback then
		if type(fallback) ~= "table" then
		    fallback = { fallback }
		end
		for _,myfnt in ipairs(fallback) do
		    local cjkfontkey = myfnt..'@'..eng.size
		    if cjk_font_list[cjkfontkey] then
			korid = cjk_font_list[cjkfontkey] 
		    else
			local size = eng.size
			---[[ size scale option
			local sizescale = string.match(myfnt,"[:;]sizescale%s*=>?%s*(%d+)")
			if tonumber(sizescale) then
			    size = size * (sizescale / 1000)
			end
			--]]
			local myfntdata = luako.fontdefine(myfnt,size,font.nextid())
			if type(myfntdata) == 'table' then
			    korid = font.define(myfntdata)
			elseif type(myfntdata) == 'number' then
			    korid = myfntdata
			end
			cjk_font_list[cjkfontkey] = korid
		    end
		    if korid then
			local f,chr = get_font_char_data(korid,curr.char)
			if chr then
			    curr.font = korid 
			    ---[[ charraise option
			    local charraise = fontdata[korid].shared
				and fontdata[korid].shared.features
				and fontdata[korid].shared.features.charraise
			    if tonumber(charraise) then
				emsize = (f and f.parameters and f.parameters.quad) or emsize
				charraise = (charraise - 1000) / 1000 * emsize
				curr.yoffset = curr.yoffset and (curr.yoffset + charraise) or charraise
			    end
			    --]]
			    ---[[ fakeslant italic correction
			    if chr.italic
				and curr.next
				and curr.next.id == kernnode
				and curr.next.kern == 0 then
				curr.next.kern = chr.italic
			    end
			    --]]
			    break
			end
		    end
		end
	    end
	    local _,ischr = get_font_char_data(korid,curr.char)
	    if not ischr then
		texio.write_nl(string.format("!Missing character: %s %d",unicode.utf8.char(curr.char),curr.char))
	    end
	end
    end
end

callback.register('hpack_filter',function(head)
    ---[[ luatexko
    korean_autojosa(head)
    remove_cj_spaceskip(head)
    font_substitute(head)
    --]]
    local tail = node.slide(head)
    head = nodes.process_characters(head,tail)
    nodes.inject_kerns(head)
    nodes.protect_glyphs(head)
    tail = node.ligaturing(head,tail)
    tail = node.kerning(head,tail)
    ---[[ luatexko
    get_ruby_side_kern(head)
    cjk_spacing_linebreak(head)
    spread_ruby_base_box(head)
    compress_fullwidth_punctuations(head)
    --no_ruby_at_margin(head)
    --]]
    return head
end)

callback.register('pre_linebreak_filter',function(head)
    ---[[ luatexko
    korean_autojosa(head)
    remove_cj_spaceskip(head)
    font_substitute(head)
    --]]
    local tail = node.slide(head)
    head = nodes.process_characters(head,tail)
    nodes.inject_kerns(head)
    nodes.protect_glyphs(head)
    tail = node.ligaturing(head,tail)
    tail = node.kerning(head,tail)
    ---[[ luatexko
    get_ruby_side_kern(head)
    cjk_spacing_linebreak(head)
    spread_ruby_base_box(head)
    compress_fullwidth_punctuations(head)
    discourage_char_widow(head, node.slide(head))
    no_ruby_at_margin(head)
    --]]
    return head
end)

--
-- dot emphasis (드러냄표 강조)
--
local function after_linebreak_dotemph (head)
    for curr in traverse(head) do
	if curr.id == hlistnode then -- hlist may be nested!!!
	    after_linebreak_dotemph(curr.list)
	elseif curr.id == glyphnode then
	    local attr = has_attribute(curr,attrdotemph)
	    if attr and attr > 0 then
		local _,chr = get_font_char_data(curr.font,curr.char)
		if chr then
		    local d = copy_node(luako.dotemphnode[attr])
		    local dot = d.list
		    d.list = get_kernnode(-chr.width/2-d.width/2)
		    d.list.next = dot
		    d.width = 0
		    insert_after(head,curr,d)
		end
		unset_attribute(curr,attrdotemph)
	    end
	end
    end
end

--
-- ruby 
--
local function after_linebreak_ruby (head)
    for curr in traverse_id(hlistnode,head) do
	after_linebreak_ruby(curr.list) -- hlist may be nested!!!
	local attr = has_attribute(curr,attrruby)
	if attr then
	    local ruby = luako.rubynode[attr][1]
	    if ruby then
		if ruby.width < curr.width then
		    local rubyhead = ruby.list
		    local numofglues = node.count(gluenode,rubyhead)
		    local extrawidth = (curr.width - ruby.width)/(numofglues + 1)
		    ruby.width = curr.width - extrawidth/2
		    if numofglues > 0 then
			spread_ruby_box(rubyhead,extrawidth)
		    end
		else
		    local right = luako.rubynode[attr].rightshift or 0
		    ruby.width = curr.width + (ruby.width-curr.width+right)/2
		end
		ruby.shift = -curr.height-luako.rubynode[attr][2]
		insert_after(head,curr,ruby)
		insert_after(head,curr,get_kernnode(-ruby.width))
	    end
	    luako.rubynode[attr] = nil
	    unset_attribute(curr,attrruby)
	end
    end
end

--
-- underline emphasis
--

local function draw_underline(head,curr,width,ulinenum)
    local glue   = get_gluenode(width)
    glue.subtype = 101 -- cleaders
    glue.leader  = copy_node(luako.ulinebox[ulinenum])
    insert_before(head,curr,get_kernnode(-width))
    insert_before(head,curr,glue)
end

local function after_linebreak_underline(head,glueorder,glueset,gluesign,ulinenum)
    local ullength = 0
    for curr in traverse(head) do
	if curr.id == whatsitnode
	    and curr.subtype == whatsitspecial
	    and curr.data then
	    if string.find(curr.data,"luako:ulinebegin=") then
		ulinenum = tonumber(string.match(curr.data,"(%d+)"))
		ullength = 0
		remove_node(head,curr)
	    elseif string.find(curr.data,'luako:ulineend') then
		if ulinenum then
		    if ullength > 0 then
			draw_underline(head,curr,ullength,ulinenum)
		    end
		    luako.ulinebox[ulinenum] = nil
		end
		remove_node(head,curr)
		ulinenum,ullength = nil,0
	    end
	elseif ulinenum then
	    if curr.id == glyphnode then
		local _,chr = get_font_char_data(curr.font,curr.char)
		local width = chr and chr.width
		ullength = ullength + (width or 0)
	    elseif curr.id == gluenode and ullength > 0 and (curr.subtype == 0 or curr.subtype > 11) and curr.spec then
		local width = curr.spec.width or 0
		if gluesign == 1 then -- see texbook p.77
		    if glueorder == curr.spec.stretch_order then
			width = width + (glueset or 0) * (curr.spec.stretch or 0)
		    end
		else
		    if glueorder == curr.spec.shrink_order then
			width = width - (glueset or 0) * (curr.spec.shrink or 0)
		    end
		end
		ullength = ullength + width
	    elseif curr.id == kernnode then
		ullength = ullength + (curr.kern or 0)
	    elseif curr.id == rulenode
		or curr.id == hlistnode
		or curr.id == vlistnode then
		ullength = ullength + (curr.width or 0)
	    elseif curr.id == marginkernnode and curr.subtype == 1 and curr.width and ullength > 0 then
		local width = curr.width
		if type(width) == "userdata" and width.id == glyphnode then 
		    local _,chr = get_font_char_data(curr.width.font,curr.width.char)
		    local rp = chr and chr.right_protruding
		    width = rp and -rp or -chr.width
		end
		ullength = ullength + (width or 0)
	    end
	    if not curr.next and ullength > 0 then
		draw_underline(head,curr,ullength,ulinenum)
	    end
	end
    end
    for curr in traverse_id(hlistnode,head) do
	local ulinenum
	after_linebreak_underline(curr.list,curr.glue_order,curr.glue_set,curr.glue_sign,ulinenum)
    end
    return ulinenum
end

callback.register("vpack_filter",function(head)
    local ulinenum
    for curr in traverse_id(hlistnode,head) do
	after_linebreak_dotemph(curr.list)
	after_linebreak_ruby(curr.list)
	ulinenum = after_linebreak_underline(curr.list,curr.glue_order,curr.glue_set,curr.glue_sign,ulinenum)
    end
    return true
end)

callback.register("post_linebreak_filter",function(head)
    local ulinenum
    for curr in traverse_id(hlistnode,head) do
	after_linebreak_dotemph(curr.list)
	after_linebreak_ruby(curr.list)
	ulinenum = after_linebreak_underline(curr.list,curr.glue_order,curr.glue_set,curr.glue_sign,ulinenum)
    end
    return true
end)

